/* $Id$

(c) Copyright, Real-Time Innovations, 2013-2016.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.


modification history
--------------------
1.0a,20feb13,acr Created
============================================================================= */

#ifndef RTI_DDS_TOPIC_TOPIC_HPP_
#define RTI_DDS_TOPIC_TOPIC_HPP_

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

#include <omg/types/string_view.hpp>

#include <dds/core/BuiltinTopicTypes.hpp>
#include <dds/core/corefwd.hpp>
#include <dds/topic/qos/TopicQos.hpp>
#include <dds/topic/TopicTraits.hpp>
#include <dds/topic/TopicListener.hpp>

#include <rti/topic/TopicDescriptionImpl.hpp>
#include <rti/core/Exception.hpp>
#include <rti/topic/detail/TopicListenerForwarder.hpp>
#include <rti/domain/PluginSupport.hpp>
#include <rti/core/detail/ListenerHolder.hpp>

namespace rti { namespace topic {

class OMG_DDS_API UntypedTopic : public virtual UntypedTopicDescription {
protected:
    ~UntypedTopic();

public:
    const dds::topic::qos::TopicQos qos() const;
    void qos(const dds::topic::qos::TopicQos& the_qos);

    dds::core::status::InconsistentTopicStatus inconsistent_topic_status() const;

    // From Entity
    void close() override
    {
        close_impl(true);
    }

    // From TopicDescription
    //
    // 'using' is to avoid partially-overridden warning:
    using UntypedTopicDescription::reserved_data;
    void reserved_data(void * ptr) override;

    // for internal use only
    DDS_Topic * native_topic()
    {
        return DDS_Topic_narrow_from_entity(this->native_entity());
    }

    // for internal use only
    const DDS_Topic * native_topic() const
    {
        return DDS_Topic_narrow_from_entity(
            (DDS_Entity *) this->native_entity());
    }

protected:
    void close_impl(bool force_close);

};

// ----------------------------------------------------------------------------

namespace detail {

using namespace rti::core::detail;

OMG_DDS_API
TopicCreationResult create_native_topic_untyped(
    const dds::domain::DomainParticipant& participant,
    const char *name,
    const char *type_name,
    const dds::topic::qos::TopicQos * qos,
    DDS_TopicListener* listener,
    const dds::core::status::StatusMask& mask);

// Returns native writer and boolean indicating if it needs to be enabled
template <typename T>
TopicCreationResult create_native_topic(
    const dds::domain::DomainParticipant& participant,
    RTI_STRING_IN name,
    RTI_STRING_IN type_name = dds::topic::topic_type_name<T>::value(),
    const dds::topic::qos::TopicQos * qos = NULL,
    dds::topic::TopicListener<T>* listener = NULL,
    const dds::core::status::StatusMask& mask =
        dds::core::status::StatusMask::all())
{
    DDS_TopicListener native_listener;
    if (listener != NULL) {
        native_listener = create_native_listener<T>(listener);
    }

    // const_cast due to issue CPPPSM-178
    dds::topic::topic_type_support<T>::register_type(
            const_cast<dds::domain::DomainParticipant&>(participant),
            std::string(type_name));

    return create_native_topic_untyped(
            participant,
            name.data(),
            type_name.data(),
            qos,
            listener ? &native_listener : NULL,
            mask);
}

// Returns native writer and boolean indicating if it needs to be enabled
OMG_DDS_API
TopicCreationResult create_native_dynamic_data_topic(
        const dds::domain::DomainParticipant& participant,
        const char *name,
        const dds::core::xtypes::DynamicType& type,
        const dds::topic::qos::TopicQos *qos = NULL,
        dds::topic::TopicListener<dds::core::xtypes::DynamicData> *listener =
                NULL,
        const dds::core::status::StatusMask& mask =
                dds::core::status::StatusMask::all());

template <typename T>
TopicCreationResult create_native_topic_no_register(
        const dds::domain::DomainParticipant& participant,
        RTI_STRING_IN the_name,
        RTI_STRING_IN type_name,
        const dds::topic::qos::TopicQos *qos = NULL,
        dds::topic::TopicListener<T> *listener = NULL,
        const dds::core::status::StatusMask& mask =
                dds::core::status::StatusMask::all())
{
    DDS_TopicListener native_listener;
    if (listener != NULL) {
        native_listener = create_native_listener<T>(listener);
    }

    return create_native_topic_untyped(
            participant,
            the_name.data(),
            type_name.data(),
            qos,
            listener ? &native_listener : NULL,
            mask);
}

// Tag type to overload the TopicImpl constructor that calls
// create_native_topic_no_register
struct no_register_tag_t {};

} // namespace detail

template <typename T>
class OMG_DDS_API TopicImpl : public TopicDescriptionImpl<T>, public UntypedTopic {
public:
    typedef dds::topic::topic_type_support<T> topic_type_support;
    typedef dds::topic::TopicListener<T> Listener;

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name)
            : UntypedTopicDescription(
                    detail::create_native_topic<T>(
                            the_participant,
                            the_name,
                            dds::topic::topic_type_name<T>::value()),
                    the_participant,
                    the_name,
                    dds::topic::topic_type_name<T>::value(),
                    detail::create_empty_listener_holder<Listener>())
    {
        // T must be a valid registered DDS type
        OMG_DDS_STATIC_ASSERT(
            (dds::topic::is_topic_type<T>::value),
            "Type T must be a DDS topic type");
    }

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name,
            RTI_STRING_IN the_type_name)
            : UntypedTopicDescription(
                    detail::create_native_topic<T>(
                            the_participant,
                            the_name,
                            the_type_name),
                    the_participant,
                    the_name,
                    the_type_name,
                    detail::create_empty_listener_holder<Listener>())
    {
        // T must be a valid registered DDS type
        OMG_DDS_STATIC_ASSERT(
            (dds::topic::is_topic_type<T>::value),
            "Type T must be a DDS topic type");
    }

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN topic_name,
            const dds::topic::qos::TopicQos& the_qos,
            dds::topic::TopicListener<T>* the_listener,
            const dds::core::status::StatusMask& mask)
            : UntypedTopicDescription(
                    detail::create_native_topic<T>(
                            the_participant,
                            topic_name,
                            dds::topic::topic_type_name<T>::value(),
                            &the_qos,
                            the_listener,
                            mask),
                    the_participant,
                    topic_name,
                    dds::topic::topic_type_name<T>::value(),
                    detail::create_empty_listener_holder<Listener>())
    {
        // T must be a valid registered DDS type
        OMG_DDS_STATIC_ASSERT(
            (dds::topic::is_topic_type<T>::value),
            "Type T must be a DDS topic type");

        if (the_listener != NULL) {
            retain_for_listener();
        }
    }

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name,
            RTI_STRING_IN the_type_name,
            const dds::topic::qos::TopicQos& the_qos,
            dds::topic::TopicListener<T> *the_listener,
            const dds::core::status::StatusMask& mask)
            : UntypedTopicDescription(
                    detail::create_native_topic<T>(
                            the_participant,
                            the_name,
                            the_type_name,
                            &the_qos,
                            the_listener,
                            mask),
                    the_participant,
                    the_name,
                    the_type_name,
                    detail::create_empty_listener_holder<Listener>())
    {
        // T must be a valid registered DDS type
        OMG_DDS_STATIC_ASSERT(
            (dds::topic::is_topic_type<T>::value),
            "Type T must be a DDS topic type");

        if (the_listener != NULL) {
            retain_for_listener();
        }
    }

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name,
            const dds::core::xtypes::DynamicType& type)
            : UntypedTopicDescription(
                    detail::create_native_dynamic_data_topic(
                            the_participant,
                            the_name.data(),
                            type),
                    the_participant,
                    the_name,
                    rti::core::xtypes::type_name(type),
                    detail::create_empty_listener_holder<Listener>())
    {
        // T must be DynamicData
        OMG_DDS_STATIC_ASSERT(
            (std::is_same<T, dds::core::xtypes::DynamicData>::value),
            "Type T must be DynamicData when creating a Topic with a DynamicType");
    }

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name,
            const dds::core::xtypes::DynamicType& type,
            const dds::topic::qos::TopicQos& the_qos,
            dds::topic::TopicListener<T> *the_listener,
            const dds::core::status::StatusMask& mask)
            : UntypedTopicDescription(
                    detail::create_native_dynamic_data_topic(
                            the_participant,
                            the_name.data(),
                            type,
                            &the_qos,
                            the_listener,
                            mask),
                    the_participant,
                    the_name,
                    rti::core::xtypes::type_name(type),
                    detail::create_empty_listener_holder<Listener>())
    {
        // T must be DynamicData
        OMG_DDS_STATIC_ASSERT(
            (std::is_same<T, dds::core::xtypes::DynamicData>::value),
            "Type T must be DynamicData when creating a Topic with a DynamicType");

        if (the_listener != NULL) {
            retain_for_listener();
        }
    }

#ifndef RTI_NO_CXX11_SMART_PTR
    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN topic_name,
            const dds::topic::qos::TopicQos& the_qos,
            std::shared_ptr<Listener> the_listener,
            const dds::core::status::StatusMask& mask)
            : UntypedTopicDescription(
                    detail::create_native_topic<T>(
                            the_participant,
                            topic_name,
                            dds::topic::topic_type_name<T>::value(),
                            &the_qos,
                            the_listener.get(),
                            mask),
                    the_participant,
                    topic_name,
                    dds::topic::topic_type_name<T>::value(),
                    detail::create_listener_holder(the_listener))
    {
        // T must be a valid registered DDS type
        OMG_DDS_STATIC_ASSERT(
                (dds::topic::is_topic_type<T>::value),
                "Type T must be a DDS topic type");
    }

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name,
            RTI_STRING_IN the_type_name,
            const dds::topic::qos::TopicQos& the_qos,
            std::shared_ptr<Listener> the_listener,
            const dds::core::status::StatusMask& mask)
            : UntypedTopicDescription(
                    detail::create_native_topic<T>(
                            the_participant,
                            the_name,
                            the_type_name,
                            &the_qos,
                            the_listener.get(),
                            mask),
                    the_participant,
                    the_name,
                    the_type_name,
                    detail::create_listener_holder(the_listener))
    {
        // T must be a valid registered DDS type
        OMG_DDS_STATIC_ASSERT(
                (dds::topic::is_topic_type<T>::value),
                "Type T must be a DDS topic type");
    }

    TopicImpl(
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name,
            const dds::core::xtypes::DynamicType& type,
            const dds::topic::qos::TopicQos& the_qos,
            std::shared_ptr<Listener> the_listener,
            const dds::core::status::StatusMask& mask)
            : UntypedTopicDescription(
                    detail::create_native_dynamic_data_topic(
                            the_participant,
                            the_name.data(),
                            type,
                            &the_qos,
                            the_listener.get(),
                            mask),
                    the_participant,
                    the_name,
                    rti::core::xtypes::type_name(type),
                    detail::create_listener_holder(the_listener))
    {
        // T must be DynamicData
        OMG_DDS_STATIC_ASSERT(
                (std::is_same<T, dds::core::xtypes::DynamicData>::value),
                "Type T must be DynamicData when creating a Topic with a "
                "DynamicType");
    }

    // This constructor doesn't register the type; it assumes that it has
    // already been registered by different means.
    //
    // For or internaly use only.)
    //
    TopicImpl(
            detail::no_register_tag_t,
            const dds::domain::DomainParticipant& the_participant,
            RTI_STRING_IN the_name,
            RTI_STRING_IN the_type_name,
            const dds::topic::qos::TopicQos *the_qos,
            std::shared_ptr<Listener> the_listener,
            const dds::core::status::StatusMask& mask)
            : UntypedTopicDescription(
                    detail::create_native_topic_no_register<T>(
                            the_participant,
                            the_name,
                            the_type_name,
                            the_qos,
                            the_listener.get(),
                            mask),
                    the_participant,
                    the_name,
                    the_type_name,
                    detail::create_listener_holder(the_listener))
    {
    }
#endif

    // for internal use
    TopicImpl(DDS_Topic* the_native_topic)
            : UntypedTopicDescription(
                    DDS_Topic_as_entity(the_native_topic),
                    DDS_Topic_as_topicdescription(the_native_topic),
                    detail::create_empty_listener_holder<Listener>())
    {
    }

    ~TopicImpl()
    {
        try {
            this->close_impl(false);
        } catch (const std::exception& ex) {
            DDSLog_logWithFunctionName(
                    RTI_LOG_BIT_EXCEPTION,
                    "~TopicImpl()",
                    &RTI_LOG_ANY_FAILURE_s,
                    ex.what());
        }
    }

    void listener_impl(
        Listener* the_listener,
        const ::dds::core::status::StatusMask& mask,
        bool retain = true)
    {
        assert_legal_listener_setter_call(retain);
        if (the_listener == NULL) {
            DDS_Topic_set_listener(native_topic(), NULL, DDS_STATUS_MASK_NONE);
            if (retain) {
                this->unretain_for_listener();
            }
        } else {
            DDS_TopicListener native_listener =
                detail::create_native_listener(the_listener);
            DDS_ReturnCode_t retcode = DDS_Topic_set_listener(
                    native_topic(),
                    &native_listener,
                    static_cast<uint32_t>(mask.to_ulong()));
            rti::core::check_return_code(retcode, "failed to set listener");

            if (retain) {
                this->retain_for_listener();
            }
        }
    }

    Listener * listener_impl() const
    {
        assert_not_closed();
        DDS_TopicListener native_listener =
            DDS_Topic_get_listener((DDS_Topic *) native_topic());

        return static_cast<Listener*>(
            native_listener.as_listener.listener_data);
    }

#ifndef RTI_NO_CXX11_SMART_PTR
    void set_listener_impl(
            std::shared_ptr<Listener> the_listener,
            const dds::core::status::StatusMask& mask)
    {
        listener_impl(the_listener.get(), mask, false);
        rti::core::detail::set_listener(listener_holder_, the_listener);
    }

    void set_listener_impl(std::shared_ptr<Listener> the_listener)
    {
        if (the_listener) {
            set_listener_impl(
                    the_listener,
                    dds::core::status::StatusMask::all());
        } else {
            set_listener_impl(nullptr, dds::core::status::StatusMask::none());
        }
    }

    std::shared_ptr<Listener> get_listener_impl() const
    {
        assert_not_closed();
        return rti::core::detail::get_listener<Listener>(listener_holder_);
    }
#endif

    // Defined here in order to avoid the 'inherits via dominance' warning
    void close()
    {
        UntypedTopic::close();
    }

    // To avoid partially-overridden warning
    using UntypedTopic::reserved_data;
    void reserved_data(void * ptr)
    {
        // Redefined to avoid inheritance-via-dominance warning
        UntypedTopic::reserved_data(ptr);
    }

};

} } // namespace rti::topic

namespace dds { namespace topic {

/** 
 * @ingroup DDSParticipantBuiltInTopicModule
 * @brief Participant builtin topic name 
 *  
 * Topic name of the builtin dds::sub::DataReader for the 
 * dds::topic::ParticipantBuiltinTopicData type 
 */
OMG_DDS_API std::string participant_topic_name();

/** 
 * @ingroup DDSTopicBuiltInTopicModule
 * @brief Topic topic name 
 *  
 * Topic name of the builtin dds::sub::DataReader for the 
 * dds::topic::TopicBuiltinTopicData type 
 */
OMG_DDS_API std::string topic_topic_name();

/** 
 * @ingroup DDSPublicationBuiltInTopicModule
 * @brief  Publication topic name 
 *  
 * Topic name of the builtin dds::sub::DataReader for the 
 * dds::topic::PublicationBuiltinTopicData type 
 */
OMG_DDS_API std::string publication_topic_name();

/** 
 * @ingroup DDSSubscriptionBuiltInTopicModule
 * @brief Subscription topic name 
 *  
 * Topic name of the builtin dds::sub::DataReader for the 
 * dds::topic::SubscriptionBuiltinTopicData type 
 */
OMG_DDS_API std::string subscription_topic_name();

} }  // namespace dds::topic

namespace rti { namespace topic {


OMG_DDS_API std::string virtual_subscription_topic_name();

/**
 * @ingroup DDSServiceRequestBuiltInTopicModule
 * @brief ServiceRequest built-in topic name
 *
 * Topic name of the builtin dds::sub::DataReader for the
 * rti::topic::ServiceRequest type
 */
OMG_DDS_API std::string service_request_topic_name();

} }  // namespace rti::topic

#endif // RTI_DDS_TOPIC_TOPIC_HPP_
