/*
 Copyright (c) 2021, Real-Time Innovations, Inc. All rights reserved.

 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.
*/

#include <dds/core/ddscore.hpp>
#include <rti/topic/cdr/GenericTypePlugin.hpp>
#include <vector>
#include <mutex>

namespace rti { namespace topic { namespace cdr {

//
// This class allows creating TypeCode and GenericTypePlugin instances and keep
// track of them so they're deleted when the factory is deleted.
//
// What this factory does is similar to what the generated code for creating
// the TypeCode with its sampleAccessInfo does, but in a dynamic way.
//
// This class is thread-safe.
//
class OMG_DDS_API GenericTypePluginFactory {
public:
    GenericTypePluginFactory() = default;
    explicit GenericTypePluginFactory(bool set_accessors);

    // Not copyable
    GenericTypePluginFactory(const GenericTypePluginFactory&) = delete;
    GenericTypePluginFactory& operator=(const GenericTypePluginFactory&) = delete;

    ~GenericTypePluginFactory();

    // Instance used to create types and plugins for serialization
    static GenericTypePluginFactory &instance();

    // Instance used to create types for public use (e.g. for DynamicData)
    static GenericTypePluginFactory &no_accessor_instance();

    // Deletes both instances if they exist
    static void delete_instance();

    dds::core::xtypes::StructType * create_struct(
            const std::string &name,
            dds::core::xtypes::ExtensibilityKind extensibility,
            uint32_t type_size,
            const std::vector<uint32_t> &member_offsets);

    dds::core::xtypes::StructType * create_struct(
            const std::string &name,
            const dds::core::xtypes::StructType &parent,
            dds::core::xtypes::ExtensibilityKind extensibility,
            uint32_t type_size,
            const std::vector<uint32_t> &member_offsets);

    dds::core::xtypes::UnionType *create_union(
            const std::string &name,
            const dds::core::xtypes::DynamicType &discriminator,
            dds::core::xtypes::ExtensibilityKind extensibility,
            uint32_t type_size,
            const std::vector<uint32_t> &member_offsets);

    dds::core::xtypes::AliasType *create_alias(
            const std::string &name,
            const dds::core::xtypes::DynamicType &aliased_type,
            uint32_t type_size);

    dds::core::xtypes::SequenceType *create_sequence(
            const dds::core::xtypes::DynamicType &element_type,
            uint32_t max_size);

    dds::core::xtypes::ArrayType *create_array(
            const dds::core::xtypes::DynamicType &element_type,
            uint32_t dimension);

    dds::core::xtypes::ArrayType *create_array(
            const dds::core::xtypes::DynamicType& element_type,
            const std::vector<uint32_t>& dimensions);

    dds::core::xtypes::EnumType *create_enum(
            const std::string &name,
            dds::core::xtypes::ExtensibilityKind extensibility_kind,
            const std::vector<dds::core::xtypes::EnumMember> &members);

    void add_member(
            dds::core::xtypes::StructType &type,
            const std::string &name,
            const dds::core::xtypes::DynamicType &member_type,
            int32_t id = dds::core::xtypes::Member::INVALID_ID,
            bool is_key = false,
            bool is_optional = false,
            bool is_external = false);

    void add_member(
            dds::core::xtypes::UnionType &type,
            const std::string &name,
            const dds::core::xtypes::DynamicType &member_type,
            const dds::core::xtypes::UnionMember::LabelSeq &labels,
            int32_t id = dds::core::xtypes::Member::INVALID_ID,
            bool is_external = false);


    CTypePlugin &create_type_plugin(
            dds::core::xtypes::DynamicType &type);

private:
    void add_type(dds::core::xtypes::DynamicType *type);
    void add_plugin(CTypePlugin *plugin);
    void assert_contains_type(const dds::core::xtypes::DynamicType &type);


    static GenericTypePluginFactory *instance_;
    static GenericTypePluginFactory *no_accessor_instance_;
    std::vector<dds::core::xtypes::DynamicType *> types_;
    std::vector<rti::topic::cdr::CTypePlugin *> plugins_;
    bool set_accessors_ = true;
    std::mutex mutex_;  // protects access to the types_, plugins_ vectors
};

} } } // namespace rti::topic::cdr