/* $Id$

(c) Copyright, Real-Time Innovations, 2022.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_DDS_TOPIC_TRUST_PARTICIPANT_TRUST_INTERCEPTOR_ALGORITHM_INFO_HPP_
#define RTI_DDS_TOPIC_TRUST_PARTICIPANT_TRUST_INTERCEPTOR_ALGORITHM_INFO_HPP_

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

#include "dds_c/dds_c_infrastructure_impl.h"

#include <dds/core/types.hpp>
#include <dds/core/Value.hpp>
#include <dds/core/Exception.hpp>
#include <rti/core/NativeValueType.hpp>

namespace rti { namespace topic { namespace trust {

class ParticipantTrustInterceptorAlgorithmInfoAdapter {
public:
    typedef struct DDS_ParticipantTrustInterceptorAlgorithmInfo native_type;

    static void initialize(native_type& native_value)
    {
        static const native_type DEFAULT =
                DDS_ParticipantTrustInterceptorAlgorithmInfo_INITIALIZER_NOSEC;
        native_value = DEFAULT;
    }

    static void finalize(native_type&)
    {
        // nothing to do
    }

    static void copy(native_type& destination, const native_type& source)
    {
        DDS_ParticipantTrustInterceptorAlgorithmInfo_copy(&destination, &source);
    }

    static bool equals(const native_type& first, const native_type& second)
    {
        return DDS_ParticipantTrustInterceptorAlgorithmInfo_equals(&first, &second)
                == DDS_BOOLEAN_TRUE;
    }
};

class ParticipantTrustInterceptorAlgorithmInfo;

} } // namespace topic::trust

namespace core {

template <>
struct native_type_traits<rti::topic::trust::ParticipantTrustInterceptorAlgorithmInfo> {
    typedef rti::topic::trust::ParticipantTrustInterceptorAlgorithmInfoAdapter adapter_type;
    typedef DDS_ParticipantTrustInterceptorAlgorithmInfo native_type;
};

}

namespace topic { namespace trust {

/**
 * @ingroup DDSBuiltInTopicCommonModule
 * @RTI_class_definition class ParticipantTrustInterceptorAlgorithmInfo
 * @headerfile rti/topic/trust/ParticipantTrustInterceptorAlgorithmInfo.hpp
 *
 * @brief @extension @st_value_type Trust Plugins interception algorithm
 * information associated with the discovered DomainParticipant.
 *
 * @details \dref_details_ParticipantTrustInterceptorAlgorithmInfo
 */
class OMG_DDS_API ParticipantTrustInterceptorAlgorithmInfo : public rti::core::NativeValueType<ParticipantTrustInterceptorAlgorithmInfo> {
public:
    RTI_NATIVE_VALUE_TYPE_DEFINE_DEFAULT_MOVE_OPERATIONS(ParticipantTrustInterceptorAlgorithmInfo)

    typedef rti::core::NativeValueType<ParticipantTrustInterceptorAlgorithmInfo> Base;
    /**
     * @brief Create an instance with the default Trust Interceptor Algorithm
     * Info associated with the discovered DomainParticipant.
     *
     * @details \dref_details_ParticipantTrustInterceptorAlgorithmInfo_new
     */
    ParticipantTrustInterceptorAlgorithmInfo() = default;

    ParticipantTrustInterceptorAlgorithmInfo(
            const struct DDS_ParticipantTrustInterceptorAlgorithmInfo& native_algo)
      : Base(native_algo)
    {
    }

    /**
     * @brief Get the Trust Plugins algorithms supported for interception of
     * data and metadata exchanged by the DomainParticipant.
     */
    uint32_t supported_mask() const
    {
        return native().supported_mask;
    }

    /**
     * @brief Get the Trust Plugins algorithms used for interception of metadata
     * exchanged by the builtin endpoints.
     *
     * @details \dref_details_ParticipantTrustInterceptorAlgorithmInfo_builtin_endpoints_required_mask
     */
    uint32_t builtin_endpoints_required_mask() const
    {
        return native().builtin_endpoints_required_mask;
    }

    /**
     * @brief Get the Trust Plugins algorithms used for interception of metadata
     * exchanged by the key exchange builtin endpoints.
     *
     * @details \dref_details_ParticipantTrustInterceptorAlgorithmInfo_builtin_kx_endpoints_required_mask
     */
    uint32_t builtin_kx_endpoints_required_mask() const
    {
        return native().builtin_kx_endpoints_required_mask;
    }

    /**
     * @brief Get the Trust Plugins algorithms used by default for
     * interception of data exchanged by the user endpoints. This is the
     * algorithm that remote DomainParticipants should assume when
     * \idref_EndpointTrustInterceptorAlgorithmInfo is not serialized.
     *
     * @details \dref_details_ParticipantTrustInterceptorAlgorithmInfo_user_endpoints_default_required_mask
     */
    uint32_t user_endpoints_default_required_mask() const
    {
        return native().user_endpoints_default_required_mask;
    }

};

} } }

#endif // RTI_DDS_TOPIC_TRUST_PARTICIPANT_TRUST_INTERCEPTOR_ALGORITHM_INFO_HPP_
