/* $Id$

(c) Copyright, Real-Time Innovations, 2021.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_UTIL_DISCOVERY_HPP_
#define RTI_UTIL_DISCOVERY_HPP_

#include <dds/sub/AnyDataReader.hpp>
#include <dds/pub/AnyDataWriter.hpp>

/** 
 * @defgroup NDDSDiscoverySnaphotModule Discovery Snapshot
 * @ingroup NDDSUtilityModule
 * @brief Utilities to print discovery snapshots of DDS entities
 */

namespace rti {
namespace util {
/**
 * @brief @extension Contains the functions to take discovery snapshot
 * for DomainParticipant, DataWriter and DataReader entities.
 */
namespace discovery {

    /**
     * @ingroup NDDSDiscoverySnaphotModule
     *
     * @brief @extension Take a snapshot of the remote participants 
     * discovered by a local one.
     *
     * @namespace_note rti::util::discovery
     * @param participant The local participant.
     * @details \dref_details_DomainParticipant_take_discovery_snapshot
     */
    OMG_DDS_API
    void take_snapshot(dds::domain::DomainParticipant participant);

    /**
     * @ingroup NDDSDiscoverySnaphotModule
     *
     * @brief @extension Take a snapshot of the remote participants 
     * discovered by a local one.
     *
     * @namespace_note rti::util::discovery
     * @param participant The local participant.
     * @param file_name Name of the file where snapshot should be printed. 
     * @details \dref_details_DomainParticipant_take_discovery_snapshot_file
     */
    OMG_DDS_API
    void take_snapshot(
            dds::domain::DomainParticipant participant,
            const std::string& file_name);

    /**
     * @ingroup NDDSDiscoverySnaphotModule
     *
     * @brief @extension Take a snapshot of the compatible and incompatible 
     * remote readers matched by a local writer.
     *
     * @namespace_note rti::util::discovery
     * @param writer The local writer.
     * @details \dref_details_DataWriter_take_discovery_snapshot
     */
    OMG_DDS_API
    void take_snapshot(dds::pub::AnyDataWriter writer);

    /**
     * @ingroup NDDSDiscoverySnaphotModule
     *
     * @brief @extension Take a snapshot of the compatible and incompatible 
     * remote readers matched by a local writer.
     *
     * @namespace_note rti::util::discovery
     * @param writer The local writer.
     * @param file_name Name of the file where snapshot should be printed.
     * @details \dref_details_DataWriter_take_discovery_snapshot_file
     */
    OMG_DDS_API
    void take_snapshot(
            dds::pub::AnyDataWriter writer,
            const std::string& file_name);

    /**
     * @ingroup NDDSDiscoverySnaphotModule
     *
     * @brief @extension Take a snapshot of the compatible and incompatible 
     * remote writers matched by a local reader.
     *
     * @namespace_note rti::util::discovery
     * @param reader Local reader.
     * @details \dref_details_DataReader_take_discovery_snapshot
     */
    OMG_DDS_API
    void take_snapshot(dds::sub::AnyDataReader reader);
    
    /**
     * @ingroup NDDSDiscoverySnaphotModule
     *
     * @brief @extension Take a snapshot of the compatible and incompatible 
     * remote writers matched by a local reader.
     *
     * @namespace_note rti::util::discovery
     * @param reader Local reader.
     * @param file_name Name of the file where snapshot should be printed.
     * @details \dref_details_DataReader_take_discovery_snapshot_file
     */
    OMG_DDS_API
    void take_snapshot(
            dds::sub::AnyDataReader reader,
            const std::string& file_name);

} } } // rti::util::discovery

#endif // RTI_UTIL_DISCOVERY_HPP_