/*
 * (c) Copyright, Real-Time Innovations, 2021.
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 */

#ifndef log_category_h
#define log_category_h

#include "log_dll.h"
#include "log_common.h"

#ifdef __cplusplus
extern "C" {
#endif

/*e \file
 * @brief Logging facilities for logging categories
 */

/*e \defgroup RTILogCategory
 * \ingroup NDDS
 */

/*
 * Size of the array of categories per instrumentBit.
 *
 * We set it to `RTI_LOG_BITMAP_MAX_PUBLIC_LOG_LEVEL + 1` to be able to map
 * directly (without any conversion) the numeric value of the instrumentBit to
 * array indexes.
 */
#define RTI_LOG_CATEGORY_MAX_SIZE (RTI_LOG_BITMAP_MAX_PUBLIC_LOG_LEVEL + 1)

/*e \ingroup RTILogCategory
 *
 * Global array of category masks. It maps from instrumentBit to categories
 * enabled in that instrumentBit. An array is used in order to have a quick
 * access to the categories, however, it is not the most efficient data
 * structure because there are many unused position in the array.
 *
 * Though the array has the number of indexes corresponding to the number of
 * bits in a verbosity definition (RTI_LOG_BITMAP_MAX_PUBLIC_LOG_LEVEL),
 * only the indexes that correspond to a possible unique log level
 * are actually populated. For example, as RTI_LOG_BIT_LOCAL equals
 * 0x00000008, the eighth index in the array is used to set the category
 * mask for LOCAL messages.
 */
extern RTILogDllVariable
RTILogCategoryMask RTILog_g_categoryMask[RTI_LOG_CATEGORY_MAX_SIZE];

/*e \ingroup RTILogCategory
 *
 * Number of RTILogCategory types. Different than the total number of
 * categories as this only refers to categories that use the category
 * filtering by thread, not the module filtering API.
 */
#define RTI_LOG_NUM_CATEGORY_MASKS 3

/*e \ingroup RTILogCategory
 *
 * Global array of labels used in the printFormat to show when you are using a
 * certain category.
 * When adding new entries to this array, the position in the array should be
 * the same as the position of the category in the RTILogCategoryBits enum.
 */
extern RTILogDllVariable
const char *RTILogCategory_g_label[RTI_LOG_NUM_CATEGORY_MASKS];

/*e
 * \ingroup RTILogCategory
 *
 * Label used for the RTI_LOG_CATEGORY_DISCOVERY_BIT category.
 */
#define RTI_LOG_CATEGORY_DISCOVERY_LABEL "Discovery"

/*e
 * \ingroup RTILogCategory
 *
 * Label used for the RTI_LOG_CATEGORY_SECURITY_BIT category.
 */
#define RTI_LOG_CATEGORY_SECURITY_LABEL "Security"

/*e
 * \ingroup RTILogCategory
 *
 * Label used for the RTI_LOG_CATEGORY_USER_BIT category.
 */
#define RTI_LOG_CATEGORY_USER_LABEL "User"

/*e
 * \ingroup RTILogCategory
 *
 * The maximum length across all labels in the RTILogCategory_g_label array.
 *
 * IMPORTANT: If you add a new label, if its length is greater than this value,
 * make sure to update it.
 */
#define RTI_LOG_CATEGORY_LABEL_MAX_LENGTH 9

/*e \ingroup RTILogCategory
 *
 * Sets the log verbosity for the given category using an integer value.
 */
extern RTILogDllExport void RTILogCategory_setCategoryVerbosity(
        RTILogCategoryMask categoryMask,
        int verbosity);

/*e \ingroup RTILogCategory
 *
 * Sets a bitmap indicating which instrumentation to use
 * for the categoryMask passed in to this function.
 */
extern RTILogDllExport void
RTILogCategory_setCategoryBitmaps(
        RTILogCategoryMask categoryMask,
        RTILogBitmap instrumentationMask);

/*e \ingroup RTILogCategory
 *
 * Gets bitmap for instrumentation being used by the input categoryMask.
 * Note that this function is intended to only be used with one category
 * enabled in the input categoryMask. Setting multiple bits in the category
 * mask will set the corresponding instrumentationMask bit to 1 if that
 * instrumentation level is set for ANY of the categories set in the
 * input categoryMask.
 */
extern RTILogDllExport void RTILogCategory_getCategoryInstrumentation(
        RTILogBitmap *instrumentationMask,
        RTILogCategoryMask categoryMask);

/* ================================================================= */
/*                         Logging APIs                              */
/* ================================================================= */

/*
 * Use these method calls to instrument the category log messages.
 */


/**
 * \ingroup RTILogCategory
 * @brief Checks verbosity level based on instrument bit,
 *      submodule and category.
 * @param moduleInstrumentBit__. \b In. Module instrumentation mask.
 * @param submoduleMask__. \b In. Submodule mask.
 * @param submoduleId__. \b In. Module ID.
 * @param instrumentBit__. \b In. Instrumentation mask.
 * @param worker__. \b In. Worker thread.
 */
#define RTILogCategory_checkVerbosity( \
        moduleInstrumentBit__, \
        submoduleMask__, \
        submoduleId__, \
        instrumentBit__, \
        worker__) \
  ((((moduleInstrumentBit__) & (instrumentBit__)) \
           && ((submoduleMask__) & (submoduleId__))) \
          || ((((worker__) != NULL) \
                      && (((const struct REDAWorker *) (worker__))->_activityContext \
                              != NULL)) \
                  && (((const struct REDAWorker *) (worker__)) \
                                  ->_activityContext->category \
                          & RTILog_g_categoryMask[(instrumentBit__)])))

/*e
 * \ingroup RTILogCategory
 * Log a message for a specific category. This compares the category set in the
 * activity context with the category in the global category mask at the chosen
 * verbosity. If these are set in both places, then the log message will be
 * logged. If the worker or activity context is NULL, we do not
 * check the activity context category and instead log the message if logging
 * is enabled for the current submodule at the specified verbosity.
 *
 * The format mask is set to RTI_LOG_PRINT_FORMAT_MASK_ALL to use the format
 * mask that the user has selected.
 *
 * This macro should generally not be directly called and you should instead use
 * the macros that are defined within each module, though directly using this
 * macro is acceptable in unit tests.
 *
 * We split this macro into two macro definitions based on whether we are
 * running cppcheck or not . This is because cppcheck will flag the cast of the
 * worker to a (struct REDAWorker*) as a nullPointer warning. We know this is a
 * false positive because we check if the worker is NULL before dereferencing
 * it. However, you cannot add a cppcheck suppression to a macro definition, we
 * would instead need to suppress it every time we call this macro. We use the
 * alternative implementation of declaring a new variable workerPtr when we are
 * running cppcheck to get around this warning. However, we do not want to
 * create a new variable on the stack every time in production.
 */
#ifdef RTI_ENABLE_CPPCHECK_SUPPRESSIONS
  #define RTILogCategory_logWithFunctionName( \
          moduleInstrumentBit__, \
          submoduleMask__, \
          submoduleId__, \
          instrumentBit__, \
          moduleId__, \
          functionName__, \
          logMessage__, \
          worker__, \
          ...) \
    { \
      struct REDAWorker *workerPtr__ = (worker__); \
      if (RTILogCategory_checkVerbosity( \
              moduleInstrumentBit__, \
              submoduleMask__, \
              submoduleId__, \
              instrumentBit__, \
              workerPtr__)) { \
        RTILogMessageParamString_printWithParams( \
                (RTI_LOG_PRINT_FORMAT_MASK_ALL), \
                (instrumentBit__), \
                (moduleId__), \
                RTI_FILE_NAME, \
                RTI_LINE_NUMBER, \
                (functionName__), \
                (logMessage__), \
                __VA_ARGS__); \
      } \
    }
#else
  #define RTILogCategory_logWithFunctionName( \
          moduleInstrumentBit__, \
          submoduleMask__, \
          submoduleId__, \
          instrumentBit__, \
          moduleId__, \
          functionName__, \
          logMessage__, \
          worker__, \
          ...) \
    if (RTILogCategory_checkVerbosity( \
            moduleInstrumentBit__, \
            submoduleMask__, \
            submoduleId__, \
            instrumentBit__, \
            worker__)) { \
      RTILogMessageParamString_printWithParams( \
              (RTI_LOG_PRINT_FORMAT_MASK_ALL), \
              (instrumentBit__), \
              (moduleId__), \
              RTI_FILE_NAME, \
              RTI_LINE_NUMBER, \
              (functionName__), \
              (logMessage__), \
              __VA_ARGS__); \
    }
#endif /* RTI_ENABLE_CPPCHECK_SUPPRESSIONS */

#define RTILogCategory_log( \
        moduleInstrumentBit__, \
        submoduleMask__, \
        submoduleId__, \
        instrumentBit__, \
        moduleId__, \
        logMessage__, \
        worker__, \
        ...) \
  RTILogCategory_logWithFunctionName( \
          (moduleInstrumentBit__), \
          (submoduleMask__), \
          (submoduleId__), \
          (instrumentBit__), \
          (moduleId__), \
          RTI_FUNCTION_NAME, \
          (logMessage__), \
          (worker__), \
          __VA_ARGS__)



#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* log_category_h */
