/* $Id$

 (c) Copyright, Real-Time Innovations, 2023
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.

=========================================================================*/

#ifndef monitoring_monitoringUtils_h
#define monitoring_monitoringUtils_h

#include "dds_common.h"
#include "monitoring/monitoring_dll.h"
#include "monitoring_dataModel.h"

#include "ndds/ndds_c.h"

#ifdef __cplusplus
    extern "C" {
#endif

/*******************************************************************************
 * Resource Tokenizer                                                          *
 ******************************************************************************/

struct RTI_MonitoringResourceTokenizer;

#define RTI_MONITORING_RESOURCE_TOKENIZER_DOUBLE_SLASH_TOKEN "$"

#define RTI_MONITORING_RESOURCE_PATH_EXPRESSION_MAX_LENGTH \
    ((size_t) (RTI_MONITORING_RESOURCE_FULLY_QUALIFIED_NAME_MAX_SIZE - 1))

#define RTI_MONITORING_RESOURCE_PATH_EXPRESSION_MAX_TOKENS 32

/*i
 * @brief Deletes a resource tokenizer.
 *
 * @param self In. Resource tokenizer.
 *
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringResourceTokenizer_delete(
        struct RTI_MonitoringResourceTokenizer *self);

/*i
 * @brief Creates a new resource tokenizer.
 *
 * The maximum length of resourcePathExpression is
 * RTI_MONITORING_RESOURCE_PATH_EXPRESSION_MAX_LENGTH.
 *
 * The resource tokenizer expands the input path expression which has the 
 * following format: [//|/]component1[//|/]component2[//|/]...[//|/]componentN
 * where componentN is a resource name, a resource name fnmatch expression, 
 * a resource GUID, or a resource class.
 *
 * For example: 
 * //data_writers/MyDataWriter
 * //data_writers/GUID(1111.1111.1111.1111)
 * //data_writers/'*'  ('' to avoid problems with C comments)
 *
 * Each component will be a token after the expression is tokenized. In 
 * addition, to make easier working with expressions there will be some other 
 * tokens automatically generated for the expression according to these rules:
 *
 * // --> Generates the token 
 * RTI_MONITORING_RESOURCE_TOKENIZER_DOUBLE_SLASH_TOKEN ($)
 * When a // is followed by a single component before the next // or the end of 
 * the expression, then the component is expanded into two tokens: *, and the 
 * component token. For example:
 *
 * The input expression //myResource will generate three tokens --> 
 * tokens are $, *, myResource
 * The input expression //myClass/myResource generates three tokens -->
 * tokens are $, myClass, myResource. There is no * token because there are 
 * two components after the //.
 * The input expression //myResource1//myResource2 generates 6 tokens -->
 * tokens are $, *, myResource1, $, *, myResource2.
 *
 * @param resourcePathExpression In. Resource path expression.
 * Cannot be NULL.
 *
 * @return New resource tokenizer or NULL if the resource path
 * expression is invalid.
 */
extern RTI_MonitoringDllExport
struct RTI_MonitoringResourceTokenizer *RTI_MonitoringResourceTokenizer_new(
        const char *resourcePathExpression);

/*i
 * @brief Set the resource path expression to tokenize.
 *
 * The maximum length of resourcePathExpression is
 * RTI_MONITORING_RESOURCE_PATH_EXPRESSION_MAX_LENGTH and the maximum number
 * of tokens is RTI_MONITORING_RESOURCE_PATH_EXPRESSION_MAX_TOKENS.
 *
 * @param self InOut. the tokenizer.
 * @param resourcePathExpression In. the resource path expression to tokenize.
 *
 * @return RTI_TRUE if the resource path expression is valid, RTI_FALSE if it
 * is invalid.
 */
extern RTI_MonitoringDllExport
RTIBool RTI_MonitoringResourceTokenizer_setExpression(
        struct RTI_MonitoringResourceTokenizer *self,
        const char *resourcePathExpression);

/*i
 * @brief Get the next token from the resource path expression
 * provided to the tokenizer constructor.
 *
 * The special token RTI_MONITORING_RESOURCE_TOKENIZER_DOUBLE_SLASH_TOKEN
 * represents two '/' characters.
 *
 * @param self InOut. the tokenizer.
 *
 * @return the next token or NULL if there are no tokens left.
 */
extern RTI_MonitoringDllExport
char * RTI_MonitoringResourceTokenizer_getNextToken(
        struct RTI_MonitoringResourceTokenizer *self);

/*i
 * @brief Get the previous token from the resource path expression
 * provided to the tokenizer constructor.
 *
 * The special token RTI_MONITORING_RESOURCE_TOKENIZER_DOUBLE_SLASH_TOKEN
 * represents two '/' characters.
 *
 * @param self InOut. the tokenizer.
 *
 * @return the previous token or NULL if the iterator was already positioned
 * at the first token or there are no tokens.
 */
extern RTI_MonitoringDllExport
char * RTI_MonitoringResourceTokenizer_getPrevToken(
        struct RTI_MonitoringResourceTokenizer *self);

/*i
 * @brief Get the first token from the resource path expression
 * provided to the tokenizer constructor.
 *
 * The special token RTI_MONITORING_RESOURCE_TOKENIZER_DOUBLE_SLASH_TOKEN
 * represents two '/' characters.
 *
 * @param self InOut. the tokenizer.
 *
 * @return the first token or NULL if there are no tokens.
 */
extern RTI_MonitoringDllExport
char * RTI_MonitoringResourceTokenizer_getFirstToken(
        struct RTI_MonitoringResourceTokenizer *self);

/*i
 * @brief Get the last token from the resource path expression
 * provided to the tokenizer constructor.
 *
 * The special token RTI_MONITORING_RESOURCE_TOKENIZER_DOUBLE_SLASH_TOKEN
 * represents two '/' characters.
 *
 * @param self InOut. the tokenizer.
 *
 * @return the last token or NULL if there are no tokens.
 */
extern RTI_MonitoringDllExport
char * RTI_MonitoringResourceTokenizer_getLastToken(
        struct RTI_MonitoringResourceTokenizer *self);

/*i
 * @brief Gets the current token.
 *
 * The special token RTI_MONITORING_RESOURCE_TOKENIZER_DOUBLE_SLASH_TOKEN
 * represents two '/' characters.
 *
 * @param self In. the tokenizer.
 *
 * @return the current token
 */
extern RTI_MonitoringDllExport
const char * RTI_MonitoringResourceTokenizer_getCurrentToken(
        const struct RTI_MonitoringResourceTokenizer *self);

/*i
 * @brief Get the numbers of tokens in the current resource path expression.
 *
 * @param self In. the tokenizer.
 *
 * @return the number of tokens.
 */
extern RTI_MonitoringDllExport
RTI_UINT32 RTI_MonitoringResourceTokenizer_getTokenCount(
        const struct RTI_MonitoringResourceTokenizer *self);

/*******************************************************************************
 * Iterator                                                                    *
 ******************************************************************************/
        
/*i
 * @brief Opaque resource used to implement
 * functionality independent of the resource registry
 * implementation.
 */
struct RTI_MonitoringOpaqueResource;

/*i
 * @brief This function prototype allows getting the next
 * sibling of a resource.
 *
 * @param resource In. Resource for which to get the next
 * sibling. Cannot be NULL.
 *
 * @return Next sibling of the resource or NULL if there is no
 * next sibling.
 */
typedef const struct RTI_MonitoringOpaqueResource *(
        *RTI_MonitoringOpaqueResource_GetNextSiblingFunction)(
        const struct RTI_MonitoringOpaqueResource *resource);

/*i
 * @brief This function prototype allows getting the first
 * child of a resource.
 *
 * @param resource In. Resource for which to get the first
 * child. Cannot be NULL.
 *
 * @return First child of the resource or NULL if there is no
 * first child.
 */
typedef const struct RTI_MonitoringOpaqueResource *(
        *RTI_MonitoringOpaqueResource_GetFirstChildFunction)(
        const struct RTI_MonitoringOpaqueResource *resource);

/*i
 * @brief This function prototype allows getting the parent
 * of a resource.
 *
 * @param resource In. Resource for which to get the parent.
 * Cannot be NULL.
 *
 * @return Parent of the resource or NULL if the resource is
 * the root resource.
 */
typedef const struct RTI_MonitoringOpaqueResource *(
        *RTI_MonitoringOpaqueResource_GetParentFunction)(
        const struct RTI_MonitoringOpaqueResource *resource);

/*i
 * @brief This function prototype allows finding a resource by
 * its GUID.
 *
 * The implementation of this function is optional. It is only there to speed up
 * the search of a resource by GUID when the expression has this form
 * //GUID(...). If this function is not implemented, the search will be done by 
 * iterating over all the resources.
 *
 * @param rootResource In. Root resource from which to start.
 * @param guid In. GUID of the resource to find. Cannot be NULL.
 *
 * @return Resource with the specified GUID or NULL if no
 * resource with the specified GUID exists.
 */
typedef const struct RTI_MonitoringOpaqueResource *(
        *RTI_MonitoringOpaqueResource_FindByGuidFunction)(
        const struct RTI_MonitoringOpaqueResource *rootResource,
        const DDS_Monitoring_GUID_t *guid);

/*i
 * @brief Opaque resource information.
 */
struct RTI_MonitoringOpaqueResourceInfo {
    char name[DDS_Monitoring_RESOURCE_NAME_LENGTH_MAX+1];
    DDS_Monitoring_GUID_t guid;
    DDS_UnsignedLong classId;
};

/*i
 * @brief This function prototype allows getting the
 * information of a resource.
 *
 * @param resource In. Resource for which to get the info. Cannot be NULL.
 * @param info Out. Resource info. Cannot be NULL.
 */
typedef void (*RTI_MonitoringOpaqueResource_GetInfoFunction)(
        const struct RTI_MonitoringOpaqueResource *resource,
        struct RTI_MonitoringOpaqueResourceInfo *info);

/*i
 * @brief This structure provides the function pointers to access the opaque 
 * resource.
 */
struct RTI_MonitoringOpaqueResourceAccessor {
    RTI_MonitoringOpaqueResource_GetNextSiblingFunction getNextSiblingFunction;
    RTI_MonitoringOpaqueResource_GetFirstChildFunction getFirstChildFunction;
    RTI_MonitoringOpaqueResource_GetParentFunction getParentFunction;
    RTI_MonitoringOpaqueResource_FindByGuidFunction findByGuidFunction;
    RTI_MonitoringOpaqueResource_GetInfoFunction getInfoFunction;
};

/*i
 * @brief RTI_MonitoringOpaqueResource
 * initializer.
 */
#define RTI_MonitoringOpaqueResourceAccessor_INITIALIZER { \
    NULL, /* getNextSiblingFunction */ \
    NULL, /* getFirstChildFunction */ \
    NULL, /* getParentFunction */ \
    NULL, /* findByGuidFunction */ \
    NULL /* getInfoFunction */ \
}

/*******************************************************************************
 * Resource Expresssion Matching                                               *
 ******************************************************************************/

DDS_SEQUENCE(
        RTI_MonitoringOpaqueResourceSeq,
        struct RTI_MonitoringOpaqueResource *);

/*i
 * @brief Returns the set of resources that match the given
 * resource path expression.
 *
 * @param resources Out. Set of resources that match the
 * resource path expression. Cannot be NULL.
 * @param resourcePathExpression In. Resource path expression.
 * Cannot be NULL.
 * @param resourceAccessor In. Resource accessor. Cannot be NULL.
 * @param rootResource In. Root resource from which to start
 * the search. Cannot be NULL.
 *
 * @return RTI_TRUE if the function succeeds, RTI_FALSE otherwise.
 */
extern RTI_MonitoringDllExport
RTIBool RTI_Monitoring_getMatchingResourcesToPathExpression(
        struct RTI_MonitoringOpaqueResourceSeq *resources,
        const char *resourcePathExpression,
        const struct RTI_MonitoringOpaqueResourceAccessor *resourceAccessor,
        const struct RTI_MonitoringOpaqueResource *rootResource);

/*i
 * @brief Returns the set of resources that match the given
 * resource path expression.
 *
 * @param resources Out. Set of resources that match the
 * resource path expression. Cannot be NULL.
 * @param resourcePathExpression In. Resource path expression.
 * Cannot be NULL.
 * @param resourceAccessor In. Resource accessor. Cannot be NULL.
 * @param rootResourceSeq In. Sequence of root resources from which to start
 * the search. Cannot be NULL.
 *
 * @return RTI_TRUE if the function succeeds, RTI_FALSE otherwise.
 */
extern RTI_MonitoringDllExport
RTIBool RTI_Monitoring_getMatchingResourcesToPathExpressionFromSequence(
        struct RTI_MonitoringOpaqueResourceSeq *resources,
        const char *resourcePathExpression,
        const struct RTI_MonitoringOpaqueResourceAccessor *resourceAccessor,
        const struct RTI_MonitoringOpaqueResourceSeq *rootResourceSeq);

/*i
 * @brief Determines if a resource matches a given resource path expression.
 *
 * @param isMatching Out. RTI_TRUE if the resource matches the expression, 
 * RTI_FALSE otherwise.
 * @param resource In. Resource to check. Cannot be NULL.
 * @param resourcePathExpression In. Resource path expression to match. 
 * Cannot be NULL.
 * @param resourceAccessor In. Resource accessor. 
 * Cannot be NULL.
 * @param rootResource In. Root resource from which to start the search. 
 * Cannot be NULL.
 *
 * @return RTI_TRUE if the function succeeds, RTI_FALSE otherwise.
 */
extern RTI_MonitoringDllExport
RTIBool RTI_Monitoring_isResourceMatchingPathExpression(
        RTIBool *isMatching,
        const struct RTI_MonitoringOpaqueResource *resource,
        const char *resourcePathExpression,
        const struct RTI_MonitoringOpaqueResourceAccessor *resourceAccessor,
        const struct RTI_MonitoringOpaqueResource *rootResource);

/*i
 * @brief Returns the set of application resources that match the given resource
 * path expression. This API is designed to find out if an expression applies to
 * an application or a set of applications. Only the application part of the
 * resource expression is checked, not the rest. Therefore, if the expression
 * referred is not an application and does not include enough information to
 * identify the application (through name or GUID), this API considers that all
 * the root resources match and returns them.
 *
 * @param resources Out. Set of application resources that match the resource 
 * path expression. Cannot be NULL.
 * @param resourcePathExpression In. Resource path expression. Cannot be NULL.
 * @param resourceAccessor In. Resource accessor. Cannot be NULL.
 * @param rootResourceSeq In. A sequence of root resources to consider in when 
 * searching. Cannot be NULL.
 *
 * @return RTI_TRUE if the function succeeds, RTI_FALSE otherwise.
 */
extern RTI_MonitoringDllExport
RTIBool RTI_Monitoring_getMatchingApplicationToPathExpressionFromSequence(
        struct RTI_MonitoringOpaqueResourceSeq *resources,
        const char *resourcePathExpression,
        const struct RTI_MonitoringOpaqueResourceAccessor *resourceAccessor,
        const struct RTI_MonitoringOpaqueResourceSeq *rootResourceSeq);

/*i
 * @brief Determines if the application part of a resource path expression
 * matches an application resource in the list of resources.
 * 
 * @param isMatching Out. RTI_TRUE if the application part of the resource path
 * expression matches an application resource in the list of resources,
 * RTI_FALSE otherwise.
 * @param resourcePathExpression In. Resource path expression to match.
 * Cannot be NULL.
 * @param resourceAccessor In. Resource accessor.
 * Cannot be NULL.
 * @param rootResource In. The root resource to match.
 * Cannot be NULL.
 * 
 * @return RTI_TRUE if the function succeeds, RTI_FALSE otherwise.
 */
RTIBool RTI_Monitoring_isMatchingApplicationToPathExpression(
        RTIBool *isMatching,
        const char *resourcePathExpression,
        const struct RTI_MonitoringOpaqueResourceAccessor *resourceAccessor,
        const struct RTI_MonitoringOpaqueResource *rootResource);

/*i
 * @brief Determines if in the given resource path expression, the last token
 * is a GUID.
 * 
 * @param resourcePathExpression In. Resource path expression to match.
 * Cannot be NULL.
 * 
 * @return RTI_TRUE if the last token is a GUID, RTI_FALSE otherwise.
 */
extern RTI_MonitoringDllExport
RTIBool RTI_Monitoring_isLastTokenGuid(const char *resourcePathExpression);

/*
 * @brief Determines if all the GUIDs in the given resource path expression
 * match the GUIDs in the list of resources.
 * 
 * @param isMatching Out. RTI_TRUE if all the GUIDs in the resource path
 * expression match the GUIDs in the list of resources, RTI_FALSE otherwise.
 * @param resourcePathExpression In. Resource path expression to match.
 * Cannot be NULL.
 * @param resourceAccessor In. Resource accessor.
 * Cannot be NULL.
 * @param firstApplicationResource In. The first application resource from which
 * to start the search.
 * Cannot be NULL.
 * 
 * @return RTI_TRUE if the function succeeds, RTI_FALSE otherwise.
 */
RTIBool RTI_Monitoring_matchAllGuidInExpression(
        RTIBool *isMatching,
        const char *resourcePathExpression,
        const struct RTI_MonitoringOpaqueResourceAccessor *resourceAccessor,
        const struct RTI_MonitoringOpaqueResource *firstApplicationResource);

#ifdef __cplusplus
    }	/* extern "C" */
#endif

#endif /* monitoring_monitoringUtils_h */
/* end of $Id$ */