/**
 * (c) Copyright, Real-Time Innovations, 2021
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 *
 * Description
 * -----------
 * RTI definitions of cryptography library structures and macros to enable
 * supporting different cryptography library versions. For example, OpenSSL
 * 1.0.2 and 1.1.1.
 * A user application does not need to include this header file. It's only
 * intended for internal use by other connextdds modules.
 * =========================================================================
 */

#ifndef security_crypto_lib_adapter_h
#define security_crypto_lib_adapter_h

#include "security/security_dll.h"
#include "dds_c/dds_c_infrastructure.h"

#ifdef RTI_WOLFSSL
  #include <wolfssl/options.h>
#endif

/* These defines are common to OpenSSL and WolfSSL (compatibility layer) */
#include <openssl/evp.h>
#include <openssl/opensslv.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/engine.h>
#include <openssl/pem.h>
#include <openssl/err.h>
#include "openssl/sha.h"
#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/stack.h>
#ifndef RTI_SECURITY_LIGHTWEIGHT
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pkcs7.h>
#endif
#include <openssl/aes.h>  /* Included for AES_BLOCK_SIZE */

extern
DDS_Boolean RTI_Security_CryptoLibAdapterAsn1TimeToSeconds(
        DDS_LongLong *seconds,
        const ASN1_TIME *asn1Time);

/***************************************************************************
 * RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE DECLARATIONS
 **************************************************************************/
/*
 * Sets custom allocators for the Crypto Library.
 *
 * @return 1 upon success, 0 if any error occurred (non-fatal).
 */
extern DDS_Boolean RTI_Security_CryptoLibAdapterUtil_setAllocators(void);

/*
 * CryptoContext stores thread-specific cryptographic state. We use the
 * CryptoContext to access TSS state withouth the need of calling (expensive)
 * TSS-APIs multiple times during cryptographic operations.
 *
 * Accessed through start() and finish() APIs.
 */
typedef struct RTI_Security_CryptoLibAdapterCryptoContext
RTI_Security_CryptoLibAdapterCryptoContext;

/**
 * @internal
 *
 * @brief This API initializes (or retrieves if it already exists) the
 * CryptoContext from the current thread's TSS and starts a CryptoContext
 * Section.
 *
 * As part of a CryptoContext Section, the implementation may perform
 * thread-specific optimizations. For example, one possible implementation of
 * this API may enable allocation optimizations for the crypto library calls
 * happening during the CryptoContext Section.
 *
 * Calls to this API must be paired with a finish() call.
 *
 * @param[in] threadContext A handle to the thread context. Can be null. If null,
 * it will be retrieved from TSS.
 * See RTI_Security_PluginSuiteUtils_getThreadContextHandle.
 *
 * @return CryptoContext from the current thread's TSS.
 */
extern RTI_Security_CryptoLibAdapterCryptoContext *
RTI_Security_CryptoLibAdapterCryptoContext_start(
        DDS_ThreadContextHandle threadContext);

/**
 * @internal
 *
 * @brief This API finishes a CryptoContext Section.
 *
 * For example, one possible implementation of this API may disable allocation
 * optimizations enabled by the start() API.
 *
 * Calls to this API must be paired with a start() call.
 *
 * @param[in] cryptoContext The CryptoContext returned by a previus call to
 * start() from the same thread. Can be null if the previous call to start()
 * from the same thread returned null.
 */
extern void RTI_Security_CryptoLibAdapterCryptoContext_finish(
        RTI_Security_CryptoLibAdapterCryptoContext *cryptoContext);

/***************************************************************************
 * SELECTED CRYPTO LIBRARY FAMILY (e.g., wolfSSL, OpenSSL)
 **************************************************************************/
#ifdef RTI_WOLFSSL
  #include "security_crypto_lib_adapter_wolfssl.h"
#else
  #include "security_crypto_lib_adapter_openssl.h"
#endif /* RTI_WOLFSSL */

/***************************************************************************
 * RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE - FALLBACK DEFS.
 **************************************************************************/
#ifndef RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
#define RTI_Security_CryptoLibAdapterUtil_setAllocators() (1)

struct RTI_Security_CryptoLibAdapterCryptoContext{
    DDS_Boolean dummyMember;
};

extern struct RTI_Security_CryptoLibAdapterCryptoContext
RTI_Security_g_dummyCryptoLibAdapterCryptoContext;


#define RTI_Security_CryptoLibAdapterCryptoContext_start(threadContext__) \
    (&RTI_Security_g_dummyCryptoLibAdapterCryptoContext); \
    RTI_Security_Utility_unusedParameter(threadContext__)
#define RTI_Security_CryptoLibAdapterCryptoContext_finish(cryptoContext__) ((void) 0)
#endif /* RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE */

/***************************************************************************
 * EVP_CIPHER_CTX - FALLBACK DEFS
 **************************************************************************/
#ifndef RTI_SUPPORT_OSSL_PARAM
  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_getGcmTag( \
          ctx__, \
          tagPointer__, /* output */ \
          tagLength__ /* input */) ( \
    EVP_CIPHER_CTX_ctrl( \
            RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_getCryptoLibCtx(ctx__), \
            EVP_CTRL_GCM_GET_TAG, \
            tagLength__, \
            tagPointer__) \
  )
  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_setGcmTag( \
          ctx__, \
          tagPointer__, /* input */ \
          tagLength__ /* input */) ( \
    EVP_CIPHER_CTX_ctrl( \
            RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_getCryptoLibCtx(ctx__), \
            EVP_CTRL_GCM_SET_TAG, \
            tagLength__, \
            tagPointer__) \
  )
#endif

#endif /* security_crypto_lib_adapter_h */