/**
 * (c) Copyright, Real-Time Innovations, 2021
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 *
 * Description
 * -----------
 * RTI definitions of cryptography library structures and macros to enable
 * supporting for OpenSSL 1.1.1 (or greater).
 * A user application does not need to include this header file. It's only
 * intended for internal use by other connextdds modules.
 * =========================================================================
 */

#ifndef security_crypto_lib_adapter_openssl111_h
#define security_crypto_lib_adapter_openssl111_h

#if !defined(RTI_WOLFSSL) \
    && (OPENSSL_VERSION_NUMBER >= RTI_OPENSSL_VERSION_1_1_0)

#ifndef RTI_SECURITY_LIGHTWEIGHT
  #define RTI_SUPPORT_OPENSSL_STORE
  #define RTI_SUPPORT_ED25519
  #define RTI_SUPPORT_ED448
  #define RTI_SUPPORT_X25519
  #define RTI_SUPPORT_X448
  #define RTI_SUPPORT_OCSP

  #include <openssl/ui.h>
  #include <openssl/store.h>

  /*
   * OpenSSL versions < 1.0.2 do not support the X509_V_FLAG_PARTIAL_CHAIN
   * flag. WolfSSL does not support it either.
   */
  #define RTI_SUPPORT_X509_V_FLAG_PARTIAL_CHAIN
#endif

  /* EVP_PKEY_cmp was deprecated by EVP_PKEY_eq in OpenSSL 3.0 */
  #ifndef RTI_Security_CryptoLibAdapterEvpPkey_equals
    #define RTI_Security_CryptoLibAdapterEvpPkey_equals(pkey1__, pkey2__) \
        (EVP_PKEY_cmp(pkey1__, pkey2__))
  #endif

    /*
     * Input/output of SSL_CTX_set_options according to
     * https://github.com/openssl/openssl/blob/OpenSSL_1_1_1t/include/openssl/ssl.h#L587
     */
  #if (OPENSSL_VERSION_NUMBER < RTI_OPENSSL_VERSION_3_0_0)
    typedef unsigned long RTI_Security_CryptoLibAdapterSslCtxOptions;
  #endif

  /***************************************************************************
   * EVP_CIPHER_CTX
   ***************************************************************************/
#ifndef RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
  #ifndef RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_CIPHER_CTX_SIZE
    /*
     * This structure is the same size as EVP_CIPHER_CTX in 64-bit. We use
     * this to declare an EVP_CIPHER_CTX on the stack instead of the heap.
     */
    #define RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_CIPHER_CTX_SIZE 168
  #endif /* RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_CIPHER_CTX_SIZE */

    typedef struct {
        /* Use 8-byte elements for alignment. */
        DDS_UnsignedLongLong dummyLongLongs[
            RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_CIPHER_CTX_SIZE / 8 + 1];
    } RTI_Security_CryptoLibAdapterEvpCipherCtxHandle;

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_getCryptoLibCtx(ctxHandle__) \
      (EVP_CIPHER_CTX *) (ctxHandle__)

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_reset(ctxHandle__) \
      EVP_CIPHER_CTX_reset((EVP_CIPHER_CTX *) (ctxHandle__))

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_initialize( \
          ctxHandle__, \
          cryptoContext__) \
      (RTI_Security_Memory_zero( \
                  (ctxHandle__), \
                  sizeof(RTI_Security_CryptoLibAdapterEvpCipherCtxHandle)), \
              (1))

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_finalize( \
          ctxHandle__, \
          cryptoContext__) \
      RTI_Security_Utility_unusedParameter(cryptoContext__); \
      EVP_CIPHER_CTX_reset((EVP_CIPHER_CTX *) (ctxHandle__))
#else
  typedef EVP_CIPHER_CTX * RTI_Security_CryptoLibAdapterEvpCipherCtxHandle;

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_setup(ctxHandle__) \
        *(ctxHandle__) = EVP_CIPHER_CTX_new()

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_cleanup(ctxHandle__) \
        EVP_CIPHER_CTX_free(*(ctxHandle__))

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_getCryptoLibCtx(ctxHandle__) \
        (*(ctxHandle__))

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_reset(ctxHandle__) \
        EVP_CIPHER_CTX_reset(*(ctxHandle__))

extern
RTI_Security_CryptoLibAdapterEvpCipherCtxHandle
RTI_Security_CryptoLibAdapterCryptoContext_getEvpCipherCtxHandle(
        RTI_Security_CryptoLibAdapterCryptoContext *cryptoContext);

extern void RTI_Security_CryptoLibAdapterCryptoContext_returnEvpCipherCtxHandle(
        RTI_Security_CryptoLibAdapterCryptoContext *cryptoContext,
        RTI_Security_CryptoLibAdapterEvpCipherCtxHandle ctxHandle);

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_initialize( \
          ctxHandle__, \
          cryptoContext__) \
      (*(ctxHandle__) =  RTI_Security_CryptoLibAdapterCryptoContext_getEvpCipherCtxHandle( \
              cryptoContext__), \
      (*(ctxHandle__) != NULL))

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_finalize( \
          ctxHandle__, \
          cryptoContext__) \
      RTI_Security_CryptoLibAdapterCryptoContext_returnEvpCipherCtxHandle( \
              cryptoContext__, \
              *(ctxHandle__))
#endif /* RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE */

  /******************************************************************************
   * EVP_MD_CTX
   ******************************************************************************/
#ifndef RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
  #ifndef RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_MD_CTX_SIZE
    /*
     * This structure is the same size as EVP_MD_CTX in 64-bit. We use this to
     * declare an EVP_MD_CTX on the stack instead of the heap.
     */
    #define RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_MD_CTX_SIZE 48
  #endif /* RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_MD_CTX_SIZE */

    typedef struct {
        /* Use 8-byte elements for alignment. */
        DDS_UnsignedLongLong dummyLongLongs[
            RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_MD_CTX_SIZE / 8 + 1];
    } RTI_Security_CryptoLibAdapterEvpMdCtxHandle;

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_getCryptoLibCtx(ctxHandle__) \
      ((EVP_MD_CTX *) ctxHandle__)

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_reset(ctxHandle__) \
      EVP_MD_CTX_reset((EVP_MD_CTX *) (ctxHandle__))

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_initialize( \
          ctxHandle__, \
          cryptoContext__) \
      (RTI_Security_Memory_zero( \
          (ctxHandle__), \
          sizeof(RTI_Security_CryptoLibAdapterEvpMdCtxHandle)), (1))
  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_finalize( \
          ctxHandle__, \
          cryptoContext__) \
      RTI_Security_Utility_unusedParameter(cryptoContext__); \
      EVP_MD_CTX_reset((EVP_MD_CTX *) (ctxHandle__))
#else
  typedef EVP_MD_CTX * RTI_Security_CryptoLibAdapterEvpMdCtxHandle;

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_setup(ctxHandle__) \
        *(ctxHandle__) = EVP_MD_CTX_new()

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_cleanup(ctxHandle__) \
        EVP_MD_CTX_free(*(ctxHandle__))

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_getCryptoLibCtx(ctxHandle__) \
        (*(ctxHandle__))

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_reset(ctxHandle__) \
        EVP_MD_CTX_reset(*(ctxHandle__))

extern
RTI_Security_CryptoLibAdapterEvpMdCtxHandle
RTI_Security_CryptoLibAdapterCryptoContext_getEvpMdCtxHandle(
        RTI_Security_CryptoLibAdapterCryptoContext *cryptoContext);

extern
void RTI_Security_CryptoLibAdapterCryptoContext_returnEvpMdCtxHandle(
        RTI_Security_CryptoLibAdapterCryptoContext *cryptoContext,
        RTI_Security_CryptoLibAdapterEvpMdCtxHandle evpMdCtx);

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_initialize( \
          ctxHandle__, \
          cryptoContext__) \
      (*(ctxHandle__) =  RTI_Security_CryptoLibAdapterCryptoContext_getEvpMdCtxHandle( \
              cryptoContext__), \
      (*(ctxHandle__) != NULL))

  #define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_finalize( \
          ctxHandle__, \
          cryptoContext__) \
      RTI_Security_CryptoLibAdapterCryptoContext_returnEvpMdCtxHandle( \
              cryptoContext__, \
              *(ctxHandle__))
#endif /* RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE */

    /***************************************************************************
     * EVP_ENCODE_CTX
     **************************************************************************/
  typedef EVP_ENCODE_CTX * RTI_Security_CryptoLibAdapterEvpEncodeCtxHandle;

  #define RTI_Security_CryptoLibAdapterEvpEncodeCtxHandle_initialize( \
          ctxHandle__) \
      (*(ctxHandle__) = EVP_ENCODE_CTX_new(), (*(ctxHandle__) != NULL))

  #define RTI_Security_CryptoLibAdapterEvpEncodeCtxHandle_finalize( \
          ctxHandle__) \
      EVP_ENCODE_CTX_free(*(ctxHandle__))

  #define RTI_Security_CryptoLibAdapterEvpEncodeCtxHandle_getCryptoLibCtx( \
          ctxHandle__) \
      (*(ctxHandle__))

  /*
   * EvpEncodeCtx does not require a separate initializer function, as
   * EVP_DecodeInit takes care of that.
   */
  /***************************************************************************
   * EVP_PKEY_new_mac_key
   **************************************************************************/
  /*
   * In OpenSSL 1.1.1, EVP_PKEY_new_mac_key() works in the same way as
   * EVP_PKEY_new_raw_private_key(). New applications should use
   * EVP_PKEY_new_raw_private_key() instead.
   */
  #define RTI_Security_CryptoLibAdapterEvpNewMacKey \
      EVP_PKEY_new_raw_private_key


  /***************************************************************************
   * RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
   **************************************************************************/
#ifdef RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
  #define RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_CRYPTO_CONTEXT_CIPHER_CTX_COUNT_MAX 2
  #define RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_CRYPTO_CONTEXT_MD_CTX_COUNT_MAX 2

    struct RTI_Security_CryptoLibAdapterCryptoContext{
        RTI_Security_CryptoLibAdapterEvpCipherCtxHandle evpCipherCtx[RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_CRYPTO_CONTEXT_CIPHER_CTX_COUNT_MAX];
        RTI_Security_CryptoLibAdapterEvpMdCtxHandle evpMdCtx[RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_CRYPTO_CONTEXT_MD_CTX_COUNT_MAX];
        struct REDAInlineMemory *threadSpecificAllocator;

        RTI_UINT32 evpCipherCtxIndex;
        RTI_UINT32 evpMdCtxIndex;
        DDS_Boolean useThreadSpecificAllocator;
    };

  #define RTI_Security_CryptoLibAdapterCryptoContext_finish(cryptoContext__) \
  { \
      if ((cryptoContext__) != NULL) { \
          (cryptoContext__)->useThreadSpecificAllocator = RTI_FALSE; \
          RTI_Security_Log_testPrecondition( \
                  (cryptoContext__)->evpCipherCtxIndex != 0, \
                  ;) \
          RTI_Security_Log_testPrecondition( \
                  (cryptoContext__)->evpMdCtxIndex != 0, \
                  ;) \
      } \
  }
#endif /* RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE */

#endif /* !defined(RTI_WOLFSSL) && (OPENSSL_VERSION_NUMBER >= RTI_OPENSSL_VERSION_1_1_0) */
#endif /* security_crypto_lib_adapter_openssl111_h */
